# Small Finite Field class (implemented with lookup tables)

# Copyright (C) 2003  Humberto Ortiz-Zuazaga
# humberto@hpcf.upr.edu
# W4-15 Los Hucares
# San Juan PR 00926-6807 USA

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

# These use bitstrings to encode fields of the form $2^k$

# Finite field class
class Field:
    def __init__(self, bitstring):
        "Uses bitstring as the generating polynomial to construct a finite field"
        self.bits = len(bitstring) - 1
        self.topbit = self.bits - 1
        self.groupsize = (2**self.bits) - 1
        self.generator = self.stringtobits(bitstring[1:])
        a = 1
        self.A = [0] * (self.groupsize)
        self.A[0] = a
        self.B = [0] * (self.groupsize+1)
        self.B[0] = 0
        self.B[1] = 0
        b = a
        for i in xrange(1,self.groupsize):
            b = self.shift(b)
            if self.isset(b,self.bits):
                b = self.unset(b, self.bits)
                b = self.plus(b, self.generator)
            self.A[i] = b
            if self.same(b,1):
                raise "polynomial %s not primitive terminated at a^%d"% (bitstring, i)
            self.B[self.A[i]] = i

    def stringtobits(self, str):
        "Read a string of '0', '1' characters, convert to a binary number, return an arbitrary precision integer with the same bit pattern"
        result = 0L
        for c in str:
            if c == '1':
                result += 1
            elif c != '0':
                raise "Invalid bit string " + str
            result *= 2
        return result/2

    def shift(self, a):
        return (a << 1)

    def plus(self, a,b):
        return (a ^ b)

    def same(self, a,b):
        return (a == b)

    def isset(self, a, pos):
        return (a & (1L << pos))

    def unset(self, a, pos):
        return (a ^ (1L << pos))

    def prod(self, a, b):
        if a == 0 or b == 0:
            return 0
        pwr = self.discretelog(a) + self.discretelog(b)
        return self.alphatothe(pwr % self.groupsize)

    def bitstostring(self, a):
        str = ""
        for i in range(self.bits):
            if self.isset(a, self.topbit - i):
                str = str + "1"
            else:
                str = str + '0'
        return str
    
    def pretty(self, a):
        if a == 0:
            return "0"
        return "a^" + str(self.discretelog(a))

    def elements(self):
        "Return all the elements of the field as a list"
        return [0] + self.A

    def multelem(self):
        "Return the elements of the multiplicative group as a list"
        return self.A

    def alphatothe(self, i):
        return self.A[i]

    def discretelog(self, a):
        return self.B[a]

    def pow(self, a,i):
        "Raise element a to the ith power"
        if a == 0:
            retval = 0
        elif i == 0:
            retval = 1
        else:
            p = 1
            retval = a
            while 2*p <= i:
                p = 2*p
                retval = self.prod(retval,retval)
            while p < i:
                p = p + 1
                retval = self.prod(retval,a)
    
        return retval

if __name__ == '__main__':

    import string
    
    # Testing, make a field
    F = Field("10000100101111")
    
    # Now do something useful
    # evaluate this poly:
    #         4653 3    1908 2    3498     5101
    # f(x) = a    x  + a    x  + a    x + a


    print "f(x) = a^4653 x^3 + a^1908 x^2  + a^3498 x + a^5101"

    def f(x):
        a3 = F.alphatothe(4653)
        x3 = F.pow(x,3)
        a2 = F.alphatothe(1908)
        x2 = F.pow(x,2)
        a1 = F.alphatothe(3498)
        a0 = F.alphatothe(5101)
        retval = F.prod(a3,x3)
        retval = F.plus(retval,F.prod(a2,x2))
        retval = F.plus(retval,F.prod(a1,x))
        retval = F.plus(retval,a0)
        return retval

    # at these points
    # S = {0*a, a^5101, a^6355, a^2550, a^3804}

    S = [0, F.alphatothe(5101), F.alphatothe(6355),
         F.alphatothe(2550), F.alphatothe(3804)]

    print "x =", map(F.pretty,S)
    
    values = map(f,S)
    print "f(x) =", map(F.pretty,values)
